/* - Spatial full -
 * This example simply displays the spatial info if it is attached and displays the acceleration data
 * for each axis as it is changing in pretty much raw form.  It also allows for modifying the sensitivity of
 * each axis that is availabl on the attached spatial.

 * Please note that this example was designed to work with only one Phidget Spatial connected. For an
 * example showing how to use two Phidgets of the same time concurrently, please see the Servo-multi example in the Servo Examples.

 * Copyright 2007 Phidgets Inc.
 * This work is licensed under the Creative Commons Attribution 2.5 Canada License.
 * To view a copy of this license, visit http://creativecommons.org/licenses/by/2.5/ca/
 */

using Phidget22;
using Phidget22.Events;
using Phidget22.ExampleUtils;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Threading;
using System.Windows.Forms;

namespace Spatial_Example {
	public partial class Form1 : Form {
		Stream _imageStream;

		CommandLineOpen openArgs;
		Spatial openCh = null;
		formCleaner cleaner;
		private Spatial spatial;
		private ErrorEventBox errorBox;

		Bitmap spatialImage;

		double[] lastTimestamp = { 0, 0, 0 };
		bool[] lastTimestampGood = { false, false, false };
		double[] gyroHeading = { 0, 0, 0 }; //degrees

		List<double[]> compassBearingFilter = new List<double[]>();
		int compassBearingFilterSize = 10;
		double compassBearing = 0;

		double ambientMagneticField = 0.57142; //Calgary
		const double ambientGravity = 1; //in G's

		public Form1() {
			InitializeComponent();
			openArgs = new CommandLineOpen(this);
			cleaner = new formCleaner(this);
		}

		public Form1(Spatial ch) {
			InitializeComponent();
			openCh = ch;
			cleaner = new formCleaner(this);
		}

		//initalize the phidget device and link the event handler code
		private void Form1_Load(object sender, EventArgs e) {

			errorBox = new ErrorEventBox(this);

			initAccelGraph();
			initGyroGraph();
			initMagFieldGraph();
			initCompassBearingGraph();

			spatial = openCh ?? openArgs.makeChannel<Spatial>();

			spatial.Attach += spatial_Attach;
			spatial.Detach += spatial_Detach;
			spatial.Error += spatial_Error;
			spatial.SpatialData += spatial_SpatialData;

			try {
				spatial.Open();
			} catch (PhidgetException ex) { errorBox.addMessage("Error opening device: " + ex.Message); }
		}

		//When the application is being terminated, close the Phidget
		private void Form1_FormClosing(object sender, FormClosingEventArgs e) {

			spatial.Attach -= spatial_Attach;
			spatial.Detach -= spatial_Detach;
			spatial.Error -= spatial_Error;
			spatial.SpatialData -= spatial_SpatialData;
			spatial.Close();
		}

		//spatial attach event handler
		void spatial_Attach(object sender, AttachEventArgs e) {
			Spatial attached = (Spatial)sender;

			Assembly _assembly = Assembly.GetExecutingAssembly();
			string imageName = _assembly.GetManifestResourceNames().Where(img => img.EndsWith("1056_0_Top.png")).FirstOrDefault(); //uses the image for the 1056.  works for all models
			_imageStream = _assembly.GetManifestResourceStream(imageName);
			spatialImage = new Bitmap(_imageStream);
			int newWidth = (int)((double)spatialImage.Width * ((double)compassView.Height / (double)spatialImage.Height) * 0.6);
			int newHeight = (int)((double)compassView.Height * 0.6);
			spatialImage = new Bitmap(spatialImage, newWidth, newWidth);

			phidgetInfoBox1.FillPhidgetInfo((Phidget)sender);

			accelBox.Visible = true;
			gyroBox.Visible = true;
			compassBox.Visible = true;
			accelGraphBox.Visible = true;
			gyroGraphBox.Visible = true;
			magFieldGraphBox.Visible = true;
			compassGraphBox.Visible = true;
			settingsBox.Visible = true;

			attached.DataInterval = 32; //setting the datarate faster than this is possible but the GUI bogs down so it isn't recommended

			dataIntervalTrk.Unit = "ms";
			dataIntervalTrk.initializeRange(attached.MinDataInterval, attached.MaxDataInterval);
			dataIntervalTrk.Value = attached.DataInterval;

			lastTimestampGood[0] = false;
			lastTimestampGood[1] = false;
			lastTimestampGood[2] = false;
			gyroHeading[0] = 0;
			gyroHeading[1] = 0;
			gyroHeading[2] = 0;
		}

		//spatial detach event handler
		void spatial_Detach(object sender, DetachEventArgs e) {
			Spatial detached = (Spatial)sender;
			phidgetInfoBox1.Clear();
			cleaner.clean();

			accelBox.Visible = false;
			gyroBox.Visible = false;
			compassBox.Visible = false;
			accelGraphBox.Visible = false;
			gyroGraphBox.Visible = false;
			magFieldGraphBox.Visible = false;
			compassGraphBox.Visible = false;
			settingsBox.Visible = false;
		}

		//spatial error event handler
		void spatial_Error(object sender, Phidget22.Events.ErrorEventArgs e) {
			errorBox.addMessage(e.Description);
		}

		void spatial_SpatialData(object sender, SpatialSpatialDataEventArgs e) {
			try {
				if (e.MagneticField[0] == (double)1e300 || e.MagneticField[1] == (double)1e300 || e.MagneticField[2] == (double)1e300)
					return;
				accelXTxt.Text = e.Acceleration[0].ToString() + " g";
				accelYTxt.Text = e.Acceleration[1].ToString() + " g";
				accelZTxt.Text = e.Acceleration[2].ToString() + " g";

				gyroXTxt.Text = e.AngularRate[0].ToString() + " /s";
				gyroYTxt.Text = e.AngularRate[1].ToString() + " /s";
				gyroZTxt.Text = e.AngularRate[2].ToString() + " /s";

				compassXTxt.Text = e.MagneticField[0].ToString() + " G";
				compassYTxt.Text = e.MagneticField[1].ToString() + " G";
				compassZTxt.Text = e.MagneticField[2].ToString() + " G";

				calculateGyroHeading(e, 0); //x axis
				calculateGyroHeading(e, 1); //y axis
				calculateGyroHeading(e, 2); //z axis

				calculateCompassBearing(e);
				drawAccelGraph(e.Acceleration);
				drawCompassBearingGraph();
				drawGyroGraph(e.AngularRate);

				drawMagFieldGraph(e.MagneticField);
			} catch (Exception ex) {
				MessageBox.Show(ex.ToString());
			}
		}

		//This finds a magnetic north bearing, correcting for board tilt and roll as measured by the accelerometer
		//This doesn't account for dynamic acceleration - ie accelerations other than gravity will throw off the calculation
		//Calculations based on AN4248
		double[] lastAngles = { 0, 0, 0 };
		void calculateCompassBearing(SpatialSpatialDataEventArgs e) {
			double[] gravity = {
				e.Acceleration[0],
				e.Acceleration[1],
				e.Acceleration[2]};

			double[] magField = {
				e.MagneticField[0],
				e.MagneticField[1],
				e.MagneticField[2]};

			//Roll Angle - about axis 0
			//  tan(roll angle) = gy/gz
			//  Use Atan2 so we have an output os (-180 - 180) degrees
			double rollAngle = Math.Atan2(gravity[1], gravity[2]);

			//Pitch Angle - about axis 1
			//  tan(pitch angle) = -gx / (gy * sin(roll angle) * gz * cos(roll angle))
			//  Pitch angle range is (-90 - 90) degrees
			double pitchAngle = Math.Atan(-gravity[0] / (gravity[1] * Math.Sin(rollAngle) + gravity[2] * Math.Cos(rollAngle)));

			//Yaw Angle - about axis 2
			//  tan(yaw angle) = (mz * sin(roll)  my * cos(roll)) /
			//                   (mx * cos(pitch) + my * sin(pitch) * sin(roll) + mz * sin(pitch) * cos(roll))
			//  Use Atan2 to get our range in (-180 - 180)
			//
			//  Yaw angle == 0 degrees when axis 0 is pointing at magnetic north
			double yawAngle = Math.Atan2(magField[2] * Math.Sin(rollAngle) - magField[1] * Math.Cos(rollAngle),
				magField[0] * Math.Cos(pitchAngle) + magField[1] * Math.Sin(pitchAngle) * Math.Sin(rollAngle) + magField[2] * Math.Sin(pitchAngle) * Math.Cos(rollAngle));

			double[] angles = { rollAngle, pitchAngle, yawAngle };

			//we low-pass filter the angle data so that it looks nicer on-screen
			try {
				//make sure the filter buffer doesn't have values passing the -180<->180 mark
				//Only for Roll and Yaw - Pitch will never have a sudden switch like that
				for (int i = 0; i < 3; i += 2) {
					if (Math.Abs(angles[i] - lastAngles[i]) > 3)
						foreach (double[] stuff in compassBearingFilter)
							if (angles[i] > lastAngles[i])
								stuff[i] += 360 * Math.PI / 180.0;
							else
								stuff[i] -= 360 * Math.PI / 180.0;
				}

				lastAngles = (double[])angles.Clone();

				compassBearingFilter.Add((double[])angles.Clone());
				if (compassBearingFilter.Count > compassBearingFilterSize)
					compassBearingFilter.RemoveAt(0);

				yawAngle = pitchAngle = rollAngle = 0;
				foreach (double[] stuff in compassBearingFilter) {
					rollAngle += stuff[0];
					pitchAngle += stuff[1];
					yawAngle += stuff[2];
				}
				yawAngle /= compassBearingFilter.Count;
				pitchAngle /= compassBearingFilter.Count;
				rollAngle /= compassBearingFilter.Count;

				//Convert radians to degrees for display
				compassBearing = yawAngle * (180.0 / Math.PI);

				bearingTxt.Text = compassBearing.ToString("F1") + "";
				pitchAngleTxt.Text = (pitchAngle * (180.0 / Math.PI)).ToString("F1") + "";
				rollAngleTxt.Text = (rollAngle * (180.0 / Math.PI)).ToString("F1") + "";
			} catch { }
		}

		private void dataIntervalTrk_KeyOverride(object sender, KeyEventArgs e) {
			e.SuppressKeyPress = true;
			dataIntervalTrk.keyOverride(sender, e);
		}

		private void slider_ValueChangeEvent(object sender, EventArgs e) {
			if (((PhidgetsControlLibrary.CustomScroll)sender).isFixedRange) //datainterval control
			{
				if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
					try {
						double temp = dataIntervalTrk.Value;
						spatial.DataInterval = (int)temp;
						if (spatial.DataInterval < (int)temp) {
							while (spatial.DataInterval != (int)temp) {
								temp++;
								spatial.DataInterval = (int)temp;
							}
						} else {
							while (spatial.DataInterval != (int)temp) {
								temp--;
								spatial.DataInterval = (int)temp;
							}
						}

						dataIntervalTrk.TrueValue = spatial.DataInterval;
						dataIntervalTrk.Value = spatial.DataInterval;
					} catch (Exception ex) { errorBox.addMessage("Error setting data interval: " + ex.Message); }
				} else if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
					try {
						spatial.DataRate = dataIntervalTrk.Value;
					} catch (Exception ex) { errorBox.addMessage("Error setting data rate: " + ex.Message); }
				}
			}
		}
		private void dataIntervalTrk_ModeSwitch(object sender, EventArgs e) {
			dataIntervalTrk.SwitchModes();
			if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
				try {
					dataIntervalTrk.isIntValue = false;
					dataIntervalTrk.Unit = "Hz";
					dataIntervalTrk.trkText = spatial.DataRate.ToString() + "Hz";
					dataIntervalTrk.Value = spatial.DataRate;
				} catch (Exception ex) { errorBox.addMessage("Error setting data rate: " + ex.Message); }

			} else if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
				try {
					dataIntervalTrk.isIntValue = true;
					if (spatial.DeviceID != DeviceID.PN_1046)
						dataIntervalTrk.Unit = "ms";
					dataIntervalTrk.trkText = spatial.DataInterval.ToString() + "ms";
					dataIntervalTrk.Value = spatial.DataInterval;
				} catch (Exception ex) { errorBox.addMessage("Error setting data interval: " + ex.Message); }
			}
			dataIntervalTrk.TrueValue = dataIntervalTrk.Value;
		}

		#region Graphs

		private BufferedGraphicsContext gyroGraphicsContext;
		private BufferedGraphics gyroGraphicsBuffer;
		private Bitmap gyroDrawingSurface;

		private Rectangle[] gyrocircleRectangle = new Rectangle[3];
		private Rectangle gyroboundsRectangle;
		private Pen[] gyroxyAxisPen = new Pen[3];
		private Pen gyrocirclePen;
		private float[] gyrocircleDiameter = new float[3], gyrocircleRadius = new float[3];

		private void initGyroGraph() {

			//initialize our graphical representation of the acceleration data
			gyroDrawingSurface = new Bitmap(gyroView.ClientRectangle.Width, gyroView.ClientRectangle.Height, System.Drawing.Imaging.PixelFormat.Format24bppRgb);

			gyroboundsRectangle = new Rectangle(0, 0, gyroDrawingSurface.Width, gyroDrawingSurface.Height);
			gyroboundsRectangle.Inflate(-1, -1);

			float xCenter = (float)(gyroboundsRectangle.Width / 2.0);
			float yCenter = (float)(gyroboundsRectangle.Height / 2.0);

			gyroxyAxisPen[0] = new Pen(Color.Red, 4);
			gyroxyAxisPen[1] = new Pen(Color.Green, 4);
			gyroxyAxisPen[2] = new Pen(Color.Blue, 4);

			gyrocirclePen = new Pen(Color.Black, 1);

			gyrocircleDiameter[0] = (float)(gyroDrawingSurface.Width / 1.5);
			gyrocircleDiameter[1] = (float)(gyroDrawingSurface.Width / 1.6);
			gyrocircleDiameter[2] = (float)(gyroDrawingSurface.Width / 1.714);

			for (int i = 0; i < 3; i++) {
				gyrocircleRadius[i] = (float)(gyrocircleDiameter[i] / 2.0);
				gyrocircleRectangle[i] = new Rectangle((int)Math.Round(xCenter - gyrocircleRadius[i]),
					(int)Math.Round(yCenter - gyrocircleRadius[i]), (int)Math.Round(gyrocircleDiameter[i]), (int)Math.Round(gyrocircleDiameter[i]));
			}
		}

		private void drawGyroGraph(double[] gyroData) {
			//get our graphics buffer going so we can draw to our panel
			gyroGraphicsContext = new BufferedGraphicsContext();
			gyroGraphicsBuffer = gyroGraphicsContext.Allocate(Graphics.FromImage(gyroDrawingSurface), gyroboundsRectangle);
			gyroGraphicsBuffer.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
			gyroGraphicsBuffer.Graphics.Clear(SystemColors.Control);


			float xCenter = (float)(gyroboundsRectangle.Width / 2.0);
			float yCenter = (float)(gyroboundsRectangle.Height / 2.0);

			//heading circles
			for (int i = 0; i < 3; i++) {
				gyroGraphicsBuffer.Graphics.DrawEllipse(gyrocirclePen, gyrocircleRectangle[i]);
				gyroGraphicsBuffer.Graphics.DrawEllipse(
					gyroxyAxisPen[i],
					(float)(gyrocircleRadius[i] * Math.Cos(gyroHeading[i] * (Math.PI / 180.0)) + xCenter) - 2,
					(float)(-gyrocircleRadius[i] * Math.Sin(gyroHeading[i] * (Math.PI / 180.0)) + yCenter) - 2,
					4,
					4
				);
			}

			gyroGraphicsBuffer.Render(gyroView.CreateGraphics());

			gyroGraphicsBuffer.Dispose();

			gyroGraphicsContext.Dispose();
		}

		private BufferedGraphicsContext accelGraphicsContext;
		private BufferedGraphics accelGraphicsBuffer;
		private Bitmap accelDrawingSurface;

		private Rectangle accelcircleRectangle, accelboundsRectangle;
		private Pen accelxyAxisPen, accelcirclePen;
		private float accelcircleDiameter, accelcircleRadius;

		private void initAccelGraph() {

			//initialize our graphical representation of the acceleration data
			accelDrawingSurface = new Bitmap(panel1.ClientRectangle.Width, panel1.ClientRectangle.Height, System.Drawing.Imaging.PixelFormat.Format24bppRgb);

			accelboundsRectangle = new Rectangle(0, 0, accelDrawingSurface.Width, accelDrawingSurface.Height);
			accelboundsRectangle.Inflate(-1, -1);

			float xCenter = (float)(accelboundsRectangle.Width / 2.0);
			float yCenter = (float)(accelboundsRectangle.Height / 2.0);

			accelxyAxisPen = new Pen(Color.Black, 2);
			accelcirclePen = new Pen(Color.DarkBlue, 2);

			accelcircleDiameter = (float)(accelDrawingSurface.Width / 1.6);
			accelcircleRadius = accelcircleDiameter / 2;

			accelcircleRectangle = new Rectangle((int)Math.Round(xCenter - accelcircleRadius), (int)Math.Round(yCenter - accelcircleRadius), (int)Math.Round(accelcircleDiameter), (int)Math.Round(accelcircleDiameter));
		}

		private void drawAccelGraph(double[] accelData) {
			accelGraphicsContext = new BufferedGraphicsContext();
			accelGraphicsBuffer = accelGraphicsContext.Allocate(Graphics.FromImage(accelDrawingSurface), accelboundsRectangle);
			accelGraphicsBuffer.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
			accelGraphicsBuffer.Graphics.Clear(SystemColors.Control);

			float xCenter = (float)(accelboundsRectangle.Width / 2.0);
			float yCenter = (float)(accelboundsRectangle.Height / 2.0);

			float xOld = xCenter - (float)accelData[0] * accelcircleRadius * (float)(1 / ambientGravity);
			float yOld = yCenter - (float)accelData[1] * accelcircleRadius * (float)(1 / ambientGravity);

			accelGraphicsBuffer.Graphics.DrawLine(accelxyAxisPen, xCenter, yCenter, xOld, yOld);
			accelGraphicsBuffer.Graphics.DrawEllipse(accelcirclePen, new Rectangle((int)(xCenter - accelcircleRadius), (int)(yCenter - accelcircleRadius),
						(int)accelcircleDiameter, (int)accelcircleDiameter));

			double zOut = accelData[2] * (float)(1 / ambientGravity);
			if (zOut > 0) {
				accelGraphicsBuffer.Graphics.DrawEllipse(new Pen(Color.Red, 2),
					new Rectangle((int)xCenter - (int)(accelcircleRadius * zOut), (int)yCenter - (int)(accelcircleRadius * zOut),
					(int)(accelcircleDiameter * zOut), (int)(accelcircleDiameter * zOut)));
			} else {
				accelGraphicsBuffer.Graphics.DrawEllipse(new Pen(Color.Green, 2),
					new Rectangle((int)xCenter - (int)(accelcircleRadius * -zOut), (int)yCenter - (int)(accelcircleRadius * -zOut),
					(int)(accelcircleDiameter * -zOut), (int)(accelcircleDiameter * -zOut)));
			}

			accelGraphicsBuffer.Render(panel1.CreateGraphics());

			accelGraphicsBuffer.Dispose();

			accelGraphicsContext.Dispose();
		}

		private BufferedGraphicsContext magFieldGraphicsContext;
		private BufferedGraphics magFieldGraphicsBuffer;
		private Bitmap magFieldDrawingSurface;

		private Rectangle magFieldCircleRectangle, magFieldBoundsRectangle;
		private Pen magFieldXYAxisPen, magFieldCirclePen;
		private float magFieldCircleDiameter, magFieldCircleRadius;

		private void initMagFieldGraph() {

			//initialize our graphical representation of the compasseration data
			magFieldDrawingSurface = new Bitmap(magFieldView.ClientRectangle.Width, magFieldView.ClientRectangle.Height, System.Drawing.Imaging.PixelFormat.Format24bppRgb);

			magFieldBoundsRectangle = new Rectangle(0, 0, magFieldDrawingSurface.Width, magFieldDrawingSurface.Height);
			magFieldBoundsRectangle.Inflate(-1, -1);

			float xCenter = (float)(magFieldBoundsRectangle.Width / 2.0);
			float yCenter = (float)(magFieldBoundsRectangle.Height / 2.0);

			magFieldXYAxisPen = new Pen(Color.Black, 2);
			magFieldCirclePen = new Pen(Color.DarkBlue, 2);

			magFieldCircleDiameter = (float)(magFieldDrawingSurface.Width / 1.6);
			magFieldCircleRadius = magFieldCircleDiameter / 2;

			magFieldCircleRectangle = new Rectangle((int)Math.Round(xCenter - magFieldCircleRadius), (int)Math.Round(yCenter - magFieldCircleRadius), (int)Math.Round(magFieldCircleDiameter), (int)Math.Round(magFieldCircleDiameter));
		}

		private void drawMagFieldGraph(double[] compassData) {
			magFieldGraphicsContext = new BufferedGraphicsContext();
			magFieldGraphicsBuffer = magFieldGraphicsContext.Allocate(Graphics.FromImage(magFieldDrawingSurface), magFieldBoundsRectangle);
			magFieldGraphicsBuffer.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
			magFieldGraphicsBuffer.Graphics.Clear(SystemColors.Control);

			float xCenter = (float)(magFieldBoundsRectangle.Width / 2.0);
			float yCenter = (float)(magFieldBoundsRectangle.Height / 2.0);

			float xOld = xCenter + (float)(compassData[0]) * magFieldCircleRadius * (float)(1 / ambientMagneticField);
			float yOld = yCenter + (float)(compassData[1]) * magFieldCircleRadius * (float)(1 / ambientMagneticField);

			magFieldGraphicsBuffer.Graphics.DrawLine(magFieldXYAxisPen, xCenter, yCenter, xOld, yOld);
			magFieldGraphicsBuffer.Graphics.DrawEllipse(magFieldCirclePen, magFieldCircleRectangle);

			double zOut = (compassData[2]) * (float)(1 / ambientMagneticField);
			if (zOut > 0) {
				magFieldGraphicsBuffer.Graphics.DrawEllipse(new Pen(Color.Red, 2),
					new Rectangle((int)xCenter - (int)(magFieldCircleRadius * zOut), (int)yCenter - (int)(magFieldCircleRadius * zOut),
					(int)(magFieldCircleDiameter * zOut), (int)(magFieldCircleDiameter * zOut)));
			} else {
				magFieldGraphicsBuffer.Graphics.DrawEllipse(new Pen(Color.Green, 2),
					new Rectangle((int)xCenter - (int)(magFieldCircleRadius * -zOut), (int)yCenter - (int)(magFieldCircleRadius * -zOut),
					(int)(magFieldCircleDiameter * -zOut), (int)(magFieldCircleDiameter * -zOut)));
			}

			magFieldGraphicsBuffer.Render(magFieldView.CreateGraphics());

			magFieldGraphicsBuffer.Dispose();

			magFieldGraphicsContext.Dispose();
		}

		private BufferedGraphicsContext compassGraphicsContext;
		private BufferedGraphics compassGraphicsBuffer;
		private Bitmap compassDrawingSurface;

		private void zeroGyroButton_Click(object sender, EventArgs e) {
			spatial.ZeroGyroscope();
			Thread.Sleep(100);
			gyroHeading[0] = 0;
			gyroHeading[1] = 0;
			gyroHeading[2] = 0;
		}

		//This integrates gyro angular rate into heading over time
		void calculateGyroHeading(SpatialSpatialDataEventArgs data, int index) {
			double gyro = 0;
			gyro = data.AngularRate[index];

			if (lastTimestampGood[index]) {
				//calculate heading
				double timechange = data.Timestamp - lastTimestamp[index]; // in ms
				double timeChangeSeconds = (double)timechange / 1000.0;

				//if (index == 0)
				//    Console.WriteLine("X Gyro rate: " + gyro.ToString() + " Time: " + timeChangeSeconds.ToString() + ", " + data[i].Timestamp.TotalMilliseconds.ToString());

				gyroHeading[index] += timeChangeSeconds * gyro;
			}

			lastTimestamp[index] = data.Timestamp;
			lastTimestampGood[index] = true;
		}

		private void setCompassParamsButton_Click(object sender, EventArgs e) {
			ParameterInputBox tib = new ParameterInputBox();
			DialogResult result = tib.ShowDialog();

			if (result == DialogResult.OK) {

				if (tib.parameters.Length == 0) {
					try {
						spatial.ResetMagnetometerCorrectionParameters();
						spatial.SaveMagnetometerCorrectionParameters();
						calibrationStatusTxt.Text = "Un-Calibrated";
						calibrationStatusTxt.ForeColor = Color.Red;
						calibrationStatusTxt.Visible = true;
					} catch (PhidgetException ex) { errorBox.addMessage("Error resetting calibration: " + ex.Message); }
					return;
				}

				string[] paramVals = tib.parameters.Split(new char[] { ',' });
				if (paramVals.Length == 13) {
					try {
						spatial.SetMagnetometerCorrectionParameters(
							double.Parse(paramVals[0]),
							double.Parse(paramVals[1]),
							double.Parse(paramVals[2]),
							double.Parse(paramVals[3]),
							double.Parse(paramVals[4]),
							double.Parse(paramVals[5]),
							double.Parse(paramVals[6]),
							double.Parse(paramVals[7]),
							double.Parse(paramVals[8]),
							double.Parse(paramVals[9]),
							double.Parse(paramVals[10]),
							double.Parse(paramVals[11]),
							double.Parse(paramVals[12])
						);
						spatial.SaveMagnetometerCorrectionParameters();
						ambientMagneticField = double.Parse(paramVals[0]);
						calibrationStatusTxt.Text = "Calibrated";
						calibrationStatusTxt.ForeColor = Color.Green;
						calibrationStatusTxt.Visible = true;
					} catch (PhidgetException ex) { errorBox.addMessage("Error calibrating: " + ex.Message); }
				}
			}
		}

		private RectangleF compasscircleRectangle;
		private Rectangle compassboundsRectangle;
		private Pen compassDotPen, compassCirclePen, compassTickPen, compassTickPenBig;
		private float compassCircleDiameter, compassCircleRadius;

		private void initCompassBearingGraph() {

			//initialize our graphical representation of the compasseration data
			compassDrawingSurface = new Bitmap(compassView.ClientRectangle.Width, compassView.ClientRectangle.Height, System.Drawing.Imaging.PixelFormat.Format24bppRgb);

			compassboundsRectangle = new Rectangle(0, 0, compassDrawingSurface.Width, compassDrawingSurface.Height);
			compassboundsRectangle.Inflate(-1, -1);

			float xCenter = (float)(compassboundsRectangle.Width / 2.0);
			float yCenter = (float)(compassboundsRectangle.Height / 2.0);

			compassDotPen = new Pen(Color.Red, 4);
			compassCirclePen = new Pen(Color.Black, 2);
			compassTickPen = new Pen(Color.DarkBlue, 1);
			compassTickPenBig = new Pen(Color.DarkBlue, 2);

			compassCircleDiameter = (float)(compassDrawingSurface.Width / 1.35);
			compassCircleRadius = (float)(compassCircleDiameter / 2.0);

			compasscircleRectangle = new RectangleF(xCenter - compassCircleRadius, yCenter - compassCircleRadius, compassCircleDiameter, compassCircleDiameter);
		}

		private void drawCompassBearingGraph() {
			compassGraphicsContext = new BufferedGraphicsContext();
			compassGraphicsBuffer = compassGraphicsContext.Allocate(Graphics.FromImage(compassDrawingSurface), compassboundsRectangle);
			compassGraphicsBuffer.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
			compassGraphicsBuffer.Graphics.Clear(SystemColors.Control);

			float xCenter = (float)(compassboundsRectangle.Width / 2.0);
			float yCenter = (float)(compassboundsRectangle.Height / 2.0);

			compassGraphicsBuffer.Graphics.DrawEllipse(compassCirclePen, compasscircleRectangle);

			Point pt = new Point((int)(xCenter - spatialImage.Width / 2), (int)(yCenter - spatialImage.Height / 2));
			compassGraphicsBuffer.Graphics.DrawImageUnscaled(rotateImage(spatialImage, (float)compassBearing), pt);

			compassGraphicsBuffer.Graphics.DrawString("N", new Font("Arial", 16), new SolidBrush(Color.Black), xCenter - 10, 0);
			compassGraphicsBuffer.Graphics.DrawString("E", new Font("Arial", 16), new SolidBrush(Color.Black), xCenter * 2 - 20, yCenter - 12);
			compassGraphicsBuffer.Graphics.DrawString("S", new Font("Arial", 16), new SolidBrush(Color.Black), xCenter - 10, yCenter * 2 - 20);
			compassGraphicsBuffer.Graphics.DrawString("W", new Font("Arial", 16), new SolidBrush(Color.Black), 0, yCenter - 12);

			//Ticks around the compass circle
			for (int i = 0; i < 360; i += 10) {
				Pen p = compassTickPen;
				int tickSize = 4;
				if (i == 0 || i == 90 || i == 180 || i == 270) {
					p = compassTickPenBig;
					tickSize = 6;
				}
				compassGraphicsBuffer.Graphics.DrawLine(p,
				(float)((compassCircleRadius + tickSize) * Math.Cos(i * (Math.PI / 180.0)) + xCenter),
				(float)(-(compassCircleRadius + tickSize) * Math.Sin(i * (Math.PI / 180.0)) + yCenter),
				(float)((compassCircleRadius - tickSize) * Math.Cos(i * (Math.PI / 180.0)) + xCenter),
				(float)(-(compassCircleRadius - tickSize) * Math.Sin(i * (Math.PI / 180.0)) + yCenter));
			}

			//Marker on the compass circle
			compassGraphicsBuffer.Graphics.DrawEllipse(
				compassDotPen,
				(float)(compassCircleRadius * Math.Cos((-compassBearing + 90) * (Math.PI / 180.0)) + xCenter) - 2,
				(float)(-compassCircleRadius * Math.Sin((-compassBearing + 90) * (Math.PI / 180.0)) + yCenter) - 2,
				4,
				4
			);

			compassGraphicsBuffer.Render(compassView.CreateGraphics());

			compassGraphicsBuffer.Dispose();

			compassGraphicsContext.Dispose();
		}

		private Bitmap rotateImage(Bitmap b, float angle) {
			//create a new empty bitmap to hold rotated image
			Bitmap returnBitmap = new Bitmap(b.Width, b.Height);
			//make a graphics object from the empty bitmap
			Graphics g = Graphics.FromImage(returnBitmap);
			//move rotation point to center of image
			g.TranslateTransform((float)b.Width / 2, (float)b.Height / 2);
			//rotate
			g.RotateTransform(angle);
			//move image back
			g.TranslateTransform(-(float)b.Width / 2, -(float)b.Height / 2);
			//draw passed in image onto graphics object
			g.DrawImage(b, new Point(0, 0));
			return returnBitmap;
		}
		#endregion
	}
}